(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  Copyright (C) 2004, 2005 Milenko "DCoder" Mitrovic                       *
 *  Mail: dcoder@dsp-worx.de                                                 *
 *  Web:  http://www.dsp-worx.de                                             *
 *                                                                           *
 *  This Program is free software; you can redistribute it and/or modify     *
 *  it under the terms of the GNU General Public License as published by     *
 *  the Free Software Foundation; either version 2, or (at your option)      *
 *  any later version.                                                       *
 *                                                                           *
 *  This Program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             *
 *  GNU General Public License for more details.                             *
 *                                                                           *
 *  You should have received a copy of the GNU General Public License        *
 *  along with GNU Make; see the file COPYING.  If not, write to             *
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.    *
 *  http://www.gnu.org/copyleft/gpl.html                                     *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)

unit DSFading;

interface

uses
  Windows, Classes, DirectSound, SysUtils, Math, DSUtil, Messages, BaseClass,
  DirectShow9;

const
  UM_PLAY   = WM_USER + 100;
  UM_SEEK   = WM_USER + 101;
  UM_ADD    = WM_USER + 102;
  UM_REMOVE = WM_USER + 103;
  UM_KILL   = WM_USER + 104;

type
  TDSPlayThread = class(TThread)
  private
    FGuid: TGuid;
    FDevice: IDirectSound;
    FCreated: Boolean;
    FList: TList;
  protected
    procedure Execute; override;
  public
    constructor Create(Guid: TGuid);
    destructor Destroy; override;
  published
    property Created: Boolean read FCreated;
    property Device:IDirectSound read FDevice write FDevice;
  end;

  TDSFading = class
  private
    function Fade: Boolean; virtual; abstract;
  end;

  TDSFadeIn = class(TDSFading)
  private
    FLastPos: Integer;
    FKilled: Boolean;
    FPlayThreadID: THandle;
    FVolume: Integer;
    FThread: THandle;
    FFadeTime: Cardinal;
    FFadeLength: Cardinal;
    FDevice: IDirectSound;
    FBuffer: IDirectSoundBuffer;
    function Fade: Boolean; override;
  public
    constructor Create(Device: IDirectSound; Buffer: IDirectSoundBuffer; FadeLength: Cardinal; Volume: Integer; PlayThreadID: THandle);
    destructor Destroy; override;
    procedure FadeIn;
    procedure Kill;
  end;

  TDSFadeOut = class(TDSFading)
  private
    FVolume: Integer;
    FThread: THandle;
    FFadeTime: Cardinal;
    FFadeLength: Cardinal;
    FDevice: IDirectSound;
    FBuffer: IDirectSoundBuffer;
    function Fade: Boolean; override;
  public
    constructor Create(Device: IDirectSound; Buffer: IDirectSoundBuffer; FadeLength: Cardinal);
    destructor Destroy; override;
    procedure FadeOut;
  end;

  procedure SetupFadeIn(Device: IDirectSound; Buffer: IDirectSoundBuffer; FadeLength: Cardinal; Volume: Integer; PlayThreadID: THandle);
  procedure SetupFadeOut(Device: IDirectSound; Buffer: IDirectSoundBuffer; FadeLength: Cardinal);

implementation

constructor TDSPlayThread.Create(Guid: TGuid);
begin
  FList := TList.Create;
  FGuid := Guid;
  FCreated := False;
  inherited Create(False);
  Priority := tpHighest;
end;

destructor TDSPlayThread.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

procedure TDSPlayThread.Execute;
var
  buffer: IDirectSoundBuffer;
  graph: IFilterGraph;
  msg: TMsg;
  p: Int64;
  i: Integer;
  seek: IMediaSeeking;
begin
  DirectSoundCreate(@FGuid, FDevice, nil);
  FCreated := True;
  
  while(GetMessage(msg, 0, 0, 0)) do
  begin
    case msg.message of
      WM_DESTROY:
      begin
        FreeOnTerminate := True;
        sleep(msg.wParam);
        while(FList.Count > 0)
          do sleep(10);
        break;
      end;
      UM_PLAY:
      begin
        Pointer(buffer) := Pointer(msg.wParam);
        buffer.Play(0, 0, DSBPLAY_LOOPING);
        buffer := nil;
        Continue;
      end;
      UM_SEEK:
      begin
        Pointer(graph) := Pointer(msg.wParam);
        seek := nil;
        graph.QueryInterface(IID_IMediaSeeking, seek);
        if Assigned(seek) then
        begin
          seek.GetCurrentPosition(p);
          seek.SetPositions(p, AM_SEEKING_AbsolutePositioning, p, AM_SEEKING_NoPositioning);
          seek := nil;
        end;
        graph := nil;
        Continue;
      end;
      UM_ADD:
      begin
        FList.Add(Pointer(msg.wParam));
      end;
      UM_REMOVE:
      begin
        FList.Remove(Pointer(msg.wParam));
      end;
      UM_KILL:
      begin
        try
          for i := 0 to FList.Count -1 do
          begin
            TDSFadeIn(FList[i]).Kill;
          end;
        except
        end;
      end;
    end;

    TranslateMessage(msg);
    DispatchMessage(msg);
  end;
end;


function ThreadProc(Fade: TDSFading): Integer; stdcall;
const
  FADETIME = 10;
begin
  Result := 0;
try
  while (Fade.Fade) do
    sleep(FADETIME);
except
end;
end;

procedure SetupFadeIn(Device: IDirectSound; Buffer: IDirectSoundBuffer; FadeLength: Cardinal; Volume: Integer; PlayThreadID: THandle);
var
  fade: TDSFadeIn;
begin
  fade := TDSFadeIn.Create(Device, Buffer, FadeLength, Volume, PlayThreadID);
  fade.FadeIn;
end;

procedure SetupFadeOut(Device: IDirectSound; Buffer: IDirectSoundBuffer; FadeLength: Cardinal);
var
  fade: TDSFadeOut;
begin
  fade := TDSFadeOut.Create(Device, Buffer, FadeLength);
  fade.FadeOut;
end;

constructor TDSFadeIn.Create(Device: IDirectSound; Buffer: IDirectSoundBuffer; FadeLength: Cardinal; Volume: Integer; PlayThreadID: THandle);
begin
  inherited Create;
  FKilled := False;
  FPlayThreadID := PlayThreadID;
  PostThreadMessage(FPlayThreadID, UM_ADD, Integer(Self), 0);
  FDevice := Device;
  FBuffer := Buffer;
  FFadeLength := FadeLength;
  FVolume := Volume;
  FLastPos := FVolume;
  FFadeTime := 0;
end;

destructor TDSFadeIn.Destroy;
begin
  FBuffer := nil;
  FDevice := nil;
  PostThreadMessage(FPlayThreadID, UM_REMOVE, Integer(Self), 0);
  inherited Destroy;
end;

procedure TDSFadeIn.FadeIn;
var
  tid: Thandle;
begin
  if (not Assigned(FBuffer)) then
  begin
    Free;
    Exit;
  end;

  FFadeTime := GetTickCount + FFadeLength;
  FThread := CreateThread(nil, 0, @ThreadProc, Self, 0, tid);
  if (FThread = 0) then
  begin
    Free;
    Exit;
  end;
  SetThreadPriority(FThread, THREAD_PRIORITY_HIGHEST);
end;

function TDSFadeIn.Fade: Boolean;
var
  time: Int64;
  vol: Integer;
  vol2: Integer;
begin
  Result := False;

  if FKilled then
  begin
    Free;
    Exit;
  end;

  time := Int64(FFadeLength) - (FFadeTime - GetTickCount);

  if time < 0
    then time := FFadeLength;

  vol2 := GetBasicAudioVolume(FVolume);
  vol := Int64(vol2) * time div FFadeLength;
  if vol > vol2
    then vol := vol2;

  vol := EnsureRange(SetBasicAudioVolume(EnsureRange(vol, 0, 10000)), -10000, 0);
  FBuffer.GetVolume(vol2);
  if (FLastPos < vol2) then
  begin
    Free;
	  Exit;
  end;
  
  FBuffer.SetVolume(vol);
  FLastPos := vol;
  
  if (time = FFadeLength) then
  begin
    Free;
    Exit;
  end;

  Result := True;
end;

procedure TDSFadeIn.Kill;
begin
  FKilled := True;
end;

constructor TDSFadeOut.Create(Device: IDirectSound; Buffer: IDirectSoundBuffer; FadeLength: Cardinal);
begin
  inherited Create;
  FDevice := Device;
  FBuffer := Buffer;
  FFadeLength := FadeLength;
  if Assigned(FBuffer)
    then FBuffer.GetVolume(FVolume)
    else FVolume := 0;
  FFadeTime := 0;
end;

destructor TDSFadeOut.Destroy;
begin
  FBuffer := nil;
  FDevice := nil;
  inherited Destroy;
end;

procedure TDSFadeOut.FadeOut;
var
  tid: Thandle;
begin
  if (FFadeLength = 0) or (not Assigned(FBuffer)) then
  begin
    Free;
    Exit;
  end;

  FFadeTime := GetTickCount + FFadeLength;

  FThread := CreateThread(nil, 0, @ThreadProc, Self, 0, tid);
  if (FThread = 0) then
  begin
    Free;
    Exit;
  end;
  SetThreadPriority(FThread, THREAD_PRIORITY_HIGHEST);
end;

function TDSFadeOut.Fade: Boolean;
var
  time: Int64;
  vol: Integer;
begin
  Result := False;

  time := FFadeTime - int64(GetTickCount);
  if (time < 0)
    then time := 0;

  vol := EnsureRange(SetBasicAudioVolume(EnsureRange(Int64(GetBasicAudioVolume(FVolume)) * time div FFadeLength, 0, 10000)), -10000, 0);
  if Assigned(FBuffer)
    then FBuffer.SetVolume(vol);

  if (time = 0) or not Assigned(FBuffer) then
  begin
    if (Assigned(FBuffer))
      then FBuffer.Stop;
    Free;
    Exit;
  end;

  Result := True;
end;

end.

